@@ -1,25 +1,27 @@
-"""Annotation."""
+"""Annotation.
+
+Core data structure to store annotations.
+"""
 import logging
 import re
 import urllib
-from typing import Dict, Optional, Tuple, Union
+from pprint import pprint
+from typing import Any, Dict, List, Optional, Tuple, Union
 
 import requests
 
 from pymetadata.core.xref import CrossReference, is_url
 from pymetadata.identifiers.miriam import BQB, BQM
-from pymetadata.identifiers.registry import Registry
+from pymetadata.identifiers.registry import REGISTRY
 from pymetadata.ontologies.ols import ONTOLOGIES, OLSQuery
 
 
-REGISTRY = Registry()
 OLS_QUERY = OLSQuery(ontologies=ONTOLOGIES)
 
-
 IDENTIFIERS_ORG_PREFIX = "https://identifiers.org"
 IDENTIFIERS_ORG_PATTERN1 = re.compile(r"^https?://identifiers.org/(.+?)/(.+)")
 IDENTIFIERS_ORG_PATTERN2 = re.compile(r"^https?://identifiers.org/(.+)")
-MIRIAM_URN_PATTERN = re.compile(r"^urn:miriam:(.+):(.+)")
+MIRIAM_URN_PATTERN = re.compile(r"^urn:miriam:(.+)")
 
 logger = logging.getLogger(__name__)
 
@@ -41,9 +43,12 @@ class RDFAnnotation:
 
     def __init__(self, qualifier: Union[BQB, BQM], resource: str):
 
+        print(qualifier, resource)
+
         self.qualifier: Union[BQB, BQM] = qualifier
         self.collection: Optional[str] = None
         self.term: Optional[str] = None
+        self.resource: str = resource
 
         if not qualifier:
             raise ValueError(
@@ -92,10 +97,13 @@ class RDFAnnotation:
         elif resource.startswith("urn:miriam:"):
             match3 = MIRIAM_URN_PATTERN.match(resource)
             if match3:
-                self.collection, self.term = match3.group(1), match3.group(2)
+
+                tokens = match3.group(1).split(":")
+                self.collection = tokens[0]
+                self.term = ":".join(tokens[1:]).replace("%3A", ":")
                 logger.warning(
                     f"Deprecated urn pattern `{resource}`, update to "
-                    f"{self.resource}"
+                    f"{self.resource_normalized}"
                 )
 
         else:
@@ -118,8 +126,18 @@ class RDFAnnotation:
                 self.collection = None
                 self.term = resource
 
+        self.clean()
+
         self.validate()
 
+    def clean(self) -> None:
+        """Clean collection and term information."""
+        if self.collection:
+            if self.collection == "biomodels.sbo":
+                self.collection = "sbo"
+            if self.collection == "obo.go":
+                self.collection = "go"
+
     @staticmethod
     def from_tuple(t: Tuple[Union[BQB, BQM], str]) -> "RDFAnnotation":
         """Construct from tuple."""
@@ -127,8 +145,14 @@ class RDFAnnotation:
         return RDFAnnotation(qualifier=qualifier, resource=resource)
 
     @property
-    def resource(self) -> Optional[str]:
-        """Resource for annotation."""
+    def resource_normalized(self) -> Optional[str]:
+        """Normalize resource for given annotation.
+
+        This is the correct usage.
+        """
+        if not self.term:
+            return None
+
         if self.collection:
             if self.term.startswith(f"{self.collection.upper()}:"):
                 return f"{IDENTIFIERS_ORG_PREFIX}/{self.term}"
@@ -137,11 +161,11 @@ class RDFAnnotation:
         else:
             return self.term
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         """Get representation string."""
         return f"RDFAnnotation({self.qualifier}|{self.collection}|{self.term})"
 
-    def to_dict(self):
+    def to_dict(self) -> Dict:
         """Convert to dict."""
         return {
             "qualifier": self.qualifier.value,  # FIXME use enums!
@@ -189,8 +213,9 @@ class RDFAnnotation:
 
     def validate(self) -> None:
         """Validate annotation."""
-        self.check_qualifier(self.qualifier)
-        if self.collection:
+        if self.qualifier:
+            self.check_qualifier(self.qualifier)
+        if self.collection and self.term:
             self.check_term(collection=self.collection, term=self.term)
 
 
@@ -203,31 +228,44 @@ class RDFAnnotationData(RDFAnnotation):
 
     def __init__(self, annotation: RDFAnnotation):
 
+        self.resource = annotation.resource
         self.qualifier = annotation.qualifier
         self.collection = annotation.collection
-        self.term = annotation.term
-        self.url: str = None
-        self.description: str = None
-        self.label: str = None
-        self.synonyms = []
-        self.xrefs = []
+        self.term: Optional[str] = annotation.term
+        self.url: Optional[str] = None
+        self.description: Optional[str] = None
+        self.label: Optional[str] = None
+        self.synonyms: List = []
+        self.xrefs: List = []
+        self.warnings: List = []
+        self.errors: List = []
 
         if self.collection:
 
             # register MIRIAM xrefs
-            namespace = REGISTRY.ns_dict.get(self.collection)
-            namespace_embedded = namespace.namespaceEmbeddedInLui
-            # print("-" * 80)
-            # print(namespace.prefix, "embedded=", namespace_embedded)
+            namespace = REGISTRY.ns_dict.get(self.collection, None)
+            if not namespace:
+                raise ValueError(
+                    f"Namespace does not exist in dict for: `{self.collection}`"
+                )
 
-            for resource in namespace.resources:  # Resource
+            namespace_embedded = namespace.namespaceEmbeddedInLui
+
+            if not namespace.resources:
+                namespace.resources = []
+
+            for ns_resource in namespace.resources:
 
                 # create url
-                url = resource.urlPattern
+                url = ns_resource.urlPattern
+
+                if not self.term:
+                    continue
+
                 term = self.term
 
                 # remove prefix
-                if namespace_embedded:
+                if namespace_embedded and namespace.prefix:
                     term = term[len(namespace.prefix) + 1 :]
 
                 # urlencode term
@@ -236,70 +274,97 @@ class RDFAnnotationData(RDFAnnotation):
                 # create url
                 url = url.replace("{$Id}", term)
                 url = url.replace("{$id}", term)
-                url = url.replace(
-                    f"{namespace.prefix.upper}:",
-                    urllib.parse.quote(f"{namespace.prefix.upper}:"),
-                )
+                if namespace.prefix:
+                    url = url.replace(
+                        f"{namespace.prefix.upper}:",
+                        urllib.parse.quote(f"{namespace.prefix.upper}:"),
+                    )
 
                 if not self.url:
                     # set url to first resource url
                     self.url = url
 
                 # print(url)
-                _xref = CrossReference(name=resource.name, accession=self.term, url=url)
-                valid = _xref.validate() and is_url(self.url)
+                _xref = CrossReference(
+                    name=ns_resource.name, accession=self.term, url=url
+                )
+                valid = _xref.validate() and is_url(self.url)  # type: ignore
                 if valid:
                     self.xrefs.append(_xref)
 
         # query OLS information
         self.query_ols()
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         """Get representation string."""
         return f"RDFAnnotationData({self.collection}|{self.term}|{self.label}|{self.description}|{self.synonyms}|{self.xrefs})"
 
-    def to_dict(self):
+    def to_dict(self) -> Dict[str, Any]:
         """Convert to dict."""
+
         return {
-            "qualifier": self.qualifier.value,
+            "resource": self.resource,
+            "resource_normalized": self.resource_normalized,
+            # "qualifier": self.qualifier.value,
             "collection": self.collection,
             "term": self.term,
-            "description": self.description,
             "label": self.label,
+            "description": self.description,
             "url": self.url,
-            # synonyms and xrefs are not serialized
+            "synonyms": self.synonyms,
+            "xrefs": self.xrefs,
+            "errors": self.errors,
+            "warnings": self.warnings,
         }
 
-    def query_ols(self):
+    def query_ols(self) -> Dict:
         """Query ontology lookup service."""
         try:
             d = OLS_QUERY.query_ols(ontology=self.collection, term=self.term)
         except requests.HTTPError as err:
             logger.error(err)
-            d = {}
+            d = {
+                "errors": [err],
+                "warnings": [],
+            }
 
         info = OLS_QUERY.process_response(d)
-        print(info)
-        self.info = info
-        if info is not None:
-            if self.label is None:
-                self.label = info.get("label")
 
-            if self.description is None:
-                self.description = info.get("description")
+        if self.label is None:
+            self.label = info["label"]
 
-            self.synonyms = info["synonyms"]
-            self.xrefs = info["xrefs"]
+        if self.description is None:
+            self.description = info["description"]
+
+        self.synonyms = info["synonyms"]
+        self.xrefs = info["xrefs"]
+        self.warnings.extend(info["warnings"])
+        self.errors.extend(info["errors"])
 
         return info
 
 
-# TODO: caching of information; fill local storage nosql
-# caching: https://realpython.com/python-memcache-efficient-caching/
-
-
 if __name__ == "__main__":
     for annotation in [
+        RDFAnnotation(
+            qualifier=BQB.IS_VERSION_OF,
+            resource="taxonomy/562",
+        ),
+        RDFAnnotation(
+            qualifier=BQB.IS_VERSION_OF,
+            resource="http://identifiers.org/taxonomy/9606",
+        ),
+        RDFAnnotation(
+            qualifier=BQB.IS_VERSION_OF,
+            resource="http://identifiers.org/biomodels.sbo/SBO:0000247",
+        ),
+        RDFAnnotation(
+            qualifier=BQB.IS_VERSION_OF, resource="urn:miriam:obo.go:GO%3A0005623"
+        ),
+        RDFAnnotation(
+            qualifier=BQB.IS_VERSION_OF, resource="urn:miriam:chebi:CHEBI%3A33699"
+        ),
+        RDFAnnotation(qualifier=BQB.IS_VERSION_OF, resource="chebi/CHEBI:456215"),
         RDFAnnotation(
             qualifier=BQB.IS, resource="https://en.wikipedia.org/wiki/Cytosol"
         ),
@@ -323,7 +388,7 @@ if __name__ == "__main__":
         RDFAnnotation(qualifier=BQB.IS_VERSION_OF, resource="BTO:0000089"),
         RDFAnnotation(qualifier=BQB.IS_VERSION_OF, resource="chebi/CHEBI:000012"),
     ]:
-        print(annotation, annotation.resource)
-
+        print("-" * 80)
         data = RDFAnnotationData(annotation)
         print(data)
+        pprint(data.to_dict())
